/*****************
	name2ip.c
		2013.9.29
		By H.Nakano
		Version 1.0
		License GPL version 3
*******************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <getopt.h>
#include <sys/socket.h>
#include <netdb.h>

void usage(){
	printf("usage: name2ip options hostname1 hostname2 ...\n");
	printf("        name2ip is a program that translates a network-hostname to the ipaddress.\n\n");
	printf("\t options\n");
	printf("\t -i,--file input file\n");
	printf("\t -o,--save output file (default: stdout)\n");
	printf("\t -e,--err  error file  (default: stderr)\n\n");
	printf("\tex)\n");
	printf("	name2ip www.kernel.org www.gnu.org\n");
	printf("	name2ip -o outfile name1 name2 ...\n");
	printf("	name2ip -i inputfile -o outfile -e errfile\n");
	printf("	name2ip -o outfile -e errfile name1 name2 ...\n");

    exit(EXIT_FAILURE);
}

static FILE *outfp = NULL;
static FILE *errfp = NULL;

void write_err(char *msg){
	errno = 0;
	int ret = fwrite((void *)msg,strlen(msg),1,errfp );
	if (ret <= 0)
		fprintf(stderr,"Error: write error-massge (%s).\n",strerror(errno));
}

static int write_dat(const char *host,const char *name){
	char line[NI_MAXHOST];
	errno = 0;
	sprintf(line,"%s  # %s\n",host,name);
	int	ret = fwrite((void *)line,strlen(line),1,outfp );
	if (ret <= 0){
		fprintf(stderr,"Error: write name(%s) (%s)\n",host,strerror(errno));
		return -1;
	 }
	return 0;
}

static int name2ip(struct gaicb **reqs,int number){
	char Error[256];
	char host[64];

	int ret = getaddrinfo_a(GAI_WAIT,reqs,number, NULL);
    if (ret != 0) {
       sprintf(Error,"Error1:%s\n",gai_strerror(ret));
		write_err(Error);
       return -1;
     }

	for(int i=0; i< number;i++){
    	ret = gai_error(reqs[i]);
    	if (ret == 0) {
    		struct addrinfo *res = reqs[i]->ar_result;
       		ret = getnameinfo(res->ai_addr, res->ai_addrlen,host,64,NULL, 0, NI_NUMERICHOST);
      		if (ret == 0){
				if (write_dat(host,reqs[i]->ar_name) < 0){
			  		ret = -1; break;
			 	}
		 	} else {
      			sprintf(Error,"Error2:(%s) %s\n",reqs[i]->ar_name,gai_strerror(ret));
				write_err(Error);
		  	}
    	 } else {
    		sprintf(Error,"Error3:(%s) %s\n",reqs[i]->ar_name,gai_strerror(ret));
			write_err(Error);
     	}
	 } 
	return 0;
}

void open_savefile(char *filename){
	if (filename == NULL) 
		outfp = stdout;
	else{
		errno = 0;
		outfp = fopen(filename,"a");
		if (outfp ==NULL){
			fprintf(stderr,"Error: open savefile(%s)(%s)\n",filename,strerror(errno));
			outfp = stdout;
		 }
	 }
}

void open_errfile(char *filename){
	if (filename == NULL) 
		errfp = stderr;
	else{
		errno = 0;
		errfp = fopen(filename,"a");
		if (errfp == NULL){
			fprintf(stderr,"Error: open errfile(%s) (%s)\n",filename,strerror(errno));
			errfp = stderr;
		 }
	 }
}

void close_file(){
	if (outfp != NULL)
		fclose(outfp);
	if (errfp != NULL)
		fclose(errfp);
}

static int get_ipaddr(char **name,int number){
	struct gaicb **reqs;
	reqs = (struct gaicb **)malloc(number* sizeof(struct gaicb*));
   
	for(int i=0;i < number;i++){
		reqs[i] = (struct gaicb *)calloc(1, sizeof(struct gaicb));
     	reqs[i]->ar_name = name[i];
	 }
	int ret = name2ip(reqs,number);

	for(int i=0;i < number;i++)
		free(reqs[i]);
	free(reqs);
	return ret;
}

#define		NMsize	 256

static int read_file(char *filename){
	char line[NI_MAXHOST];
	char *p,*dat;
	char **name = NULL;
	FILE *fp;

	errno = 0;
	fp = fopen(filename,"r");
	if (fp ==NULL){
		fprintf(stderr,"Error: open (%s) (%s)\n",filename,strerror(errno));
		return -1;
	 }

	int num = 1;
	int cc = 0;
	errno = 0;
	name = (char **)malloc(NMsize*sizeof(char *));
	if (name == NULL){
		fprintf(stderr,"Error: malloc\n");
		return -1;
	 }

	while(fgets(line,NI_MAXHOST,fp) != NULL) {
		p = line;
		while(isspace(*p)) p++;
		if (p[0] == '#' || strlen(p) <= 3) 
		   continue;

		if (cc >= NMsize*num){
			num ++;
			errno = 0;
			name = realloc(name,NMsize*num*sizeof(char *));
			if (name == NULL){
				fprintf(stderr,"Error: realloc (%s)\n",strerror(errno));
				cc = -1;
				break;
			 }
		 }

		dat = p;
		while(! isspace(*p)) p++;
 		*p = '\0';
		name[cc]= (char *)malloc(strlen(dat)*sizeof(char));
		strcpy(name[cc],dat);
		cc++;
	 }
	fclose(fp);

	int ret = 0;
	if (cc >= 1){
		if(get_ipaddr(name,cc) < 0)
			ret = -1;
		for(int i=0;i < cc;i++)
	 	   free(name[i]);
	 } else
		ret = -1;
	if(name != NULL)
		free(name);
	return ret;
}

int main(int argc, char *argv[]) {

	if (argc == 1)
		usage();

	const struct option longopt[] = {
    	{"help", no_argument, NULL, 'h'},
    	{"file", required_argument, NULL, 'i'},
  		{"save", required_argument, NULL, 'o'},
  		{"err",  required_argument, NULL, 'e'},
    	{NULL, 0, NULL, 0}
     };

  	int  c;
	char *infile = NULL;
	char *savefile = NULL;
	char *errfile  = NULL;
	
  	opterr = 0;
 	while ((c = getopt_long(argc, argv, "i:o:e:",longopt,NULL)) != -1){
     switch (c)
       {
     	case 'i': /* file */
			infile = optarg;
        	break;
      	case 'o': /* save */
			savefile = optarg;
        	break;
      	case 'e': /* error */
			errfile = optarg;
        	break;
      	default: 
        	usage();
			break;
       }
	}

	open_savefile(savefile);
	open_errfile(errfile);
	int ret = 0;
	if (infile != NULL)
		ret = read_file(infile);
	 
	if (optind < argc) {
		ret = get_ipaddr( &argv[optind],argc - optind);
	 }

	close_file();
	return ret;
}

